__all__ = [
    'ENV_VARS',
    'DEFAULT_MAX_RETRIES',
    'DEFAULT_RETRY_INTERVAL',
    'DEFAULT_TIMEOUT_IN_MS',
    'RETRY_RANDOMIZATION_FACTOR',
    'DEFAULT_LOGGER_FORMAT',
    'DEFAULT_LOGGER_DATEFMT',
    'SPARK_SDK',
]

DEFAULT_MAX_RETRIES = 2
DEFAULT_RETRY_INTERVAL = 1  # 1 second
DEFAULT_TIMEOUT_IN_MS = 60000  # 60 seconds
RETRY_RANDOMIZATION_FACTOR = 1.5
DEFAULT_LOGGER_FORMAT = '[%(name)s] %(asctime)s %(levelname)s - %(message)s'
DEFAULT_LOGGER_DATEFMT = '%m/%d/%Y, %I:%M:%S %p'
SPARK_SDK = 'Spark Python SDK'


class ENV_VARS:
    BASE_URL = 'CSPARK_BASE_URL'
    API_KEY = 'CSPARK_API_KEY'
    BEARER_TOKEN = 'CSPARK_BEARER_TOKEN'
    CLIENT_ID = 'CSPARK_CLIENT_ID'
    CLIENT_SECRET = 'CSPARK_CLIENT_SECRET'
    OAUTH_PATH = 'CSPARK_OAUTH_PATH'
